%{
//---------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                   (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//  DssLexer.l: --- lexical analysis of DSS-compliant expression ---
//
//  DESCRIPTION:
//      This is a flex/lex input file for generating a scanner for the
//      lexical analysis of boolean expressions to be converted into
//      a DataSubspace object.
//      If flex is to be used, the macro
//          FLEX_SCANNER
//      need to be defined which triggers the usage of streamlined I/O
//      routines.
//
//  CHANGE HISTORY:
//      UL, 1999-11-16: 1.0.0: - first version
//
//  AUTHOR:
//      U. Lammers (UL), Tue Nov 16 16:05:02 MET 1999
//
//  $Header: /sasbuild/repositories/sasdev/sas/packages/dsslib/src/build/DssLexer.l,v 1.1.1.1 2003/11/24 16:10:37 eojero Exp $
//

//#include <values.h> // This may make Debian Linux happy
#include <cmath>
#include <cassert>
#include <cstring>
#include <cstdlib>
#include <string>

#include "basicdefs.h"
#include "DssBuilder.h"
namespace DssParser {
    enum RelOp {LtOpT=1, LeOpT, GtOpT, GeOpT, EqOpT, NeOpT};
    #include "DssParser.h"              //  symbols form the grammar
}
#include "DssMisc.h"
#include "SetTableColumn.h"
#include "SelectError.h"

namespace DssParser {

#if defined(FLEX_SCANNER)           //  we use flex
    #undef  YY_INPUT
    #define YY_INPUT(buf,result,max_size) \
                result = DssParser::lexiread(buf, max_size);
    static int lexiread (char *, int);
#else
    #undef  input
    #undef  unput
    #undef  output
    static int input (void), unput (int), output (int);
#endif

#define dsswrap()   1               // avoid usage of liblex
}

%}

a       [aA]
b       [bB]
c       [cC]
d       [dD]
e       [eE]
f       [fF]
g       [gG]
h       [hH]
i       [iI]
k       [kK]
l       [lL]
m       [mM]
n       [nN]
o       [oO]
p       [pP]
q       [qQ]
r       [rR]
s       [sS]
t       [tT]
u       [uU]
w       [wW]
x       [xX]
y       [yY]

white   [ \t\n]
alpha   [a-zA-Z_]
alnum   [a-zA-Z_0-9]
/* alphanumeric + dot (.) and dash (-) */
alnumx  [a-zA-Z_0-9\.\-]

    /* ------------( binary, octal, decimal, hexadecimal digits )--------*/
bdigit  [01]
odigit  [0-7]
ddigit  [0-9]
hdigit  [0-9aAbBcCdDeEfF]

    /* ---------(binary, octal, decimal, hexadecimal, real numbers )------*/
binnum  ({b}{bdigit}+)
octnum  (o{odigit}+)
decnum  ({ddigit}+)
hexnum  (0{x}{hdigit}+)
exp     ({e}[+-]?{ddigit}+)
realnum ((({ddigit}+\.?{ddigit}*)|({ddigit}*\.{ddigit}+)){exp}?)

    /* ---------------( true, false, ifthenelse, selected )--------------*/
true    ({t}{r}{u}{e})
false   ({f}{a}{l}{s}{e})
ifthenelse  ({i}{f}{t}{h}{e}{n}{e}{l}{s}{e})
selected    ({s}{e}{l}{e}{c}{t}{e}{d})

    /* -----------------( arithmetic functions )-------------------------*/
fabs    ({a}{b}{s})

fsin    ({s}{i}{n})
fcos    ({c}{o}{s})
ftan    ({t}{a}{n})
arc     ({a}{r}{c})
asin    ({arc}{fsin})
acos    ({arc}{fcos})
atan    ({arc}{ftan})
atan2   ({arc}{ftan}"2")

cosh    ({fcos}{h})
sinh    ({fsin}{h})
tanh    ({ftan}{h})

fexp    ({e}{x}{p})
flog    (({l}{o}{g})|({l}{n}))
flog10  ({flog}"10")
fsqrt   ({s}{q}{r}{t})

modf    ({m}{o}{d}{f})
pow     ({p}{o}{w})
ceil    ({c}{e}{i}{l})
floor   ({f}{l}{o}{o}{r})
fmod    ({f}{m}{o}{d})
fint    ({i}{n}{t})

point   ({p}{o}{i}{n}{t})
line    ({l}{i}{n}{e})
box     ({b}{o}{x})
rect    ({r}{e}{c}{t}{a}{n}{g}{l}{e})
diamond (({d}{i}{a}{m}{o}{n}{d})|({r}{h}{o}{m}{b}{u}{s}))
sector  (({p}{i}{e})|({s}{e}{c}{t}{o}{r}))
circle  ({c}{i}{r}{c}{l}{e})
ring    (({r}{i}{n}{g})|({a}{n}{n}{u}{l}{u}{s}))
ellip   ({e}{l}{l}{i}{p})
ellipse ({ellip}{s}{e})
elliptr ({ellip}{t}{ring})
polygon ({p}{o}{l}{y}{g}{o}{n})
region  ({r}{e}{g}{i}{o}{n})
near    ({n}{e}{a}{r})

/* optional whitespace followed by '(' */
wpar    ({white}*"(")

in      ({i}{n})
gti     ({g}{t}{i})
mask    ({m}{a}{s}{k})
path    ((\/?{alnumx}*)*\/?)
/* extensiom number */
extspec ((\+{alnumx}+)|(\[{alnumx}+\])|(:{alnumx}+))
/* file name */
blockname   ({path}{alnumx}*{extspec}?)
/* attribute (=header keyword) name */
attname ([a-zA-Z][0-9a-zA-Z_\-]*)

%s EXPECTBLOCKNAME

%%
{white}{in}{white}      return IN;              /* IN operator              */
{white}+                ;                       /* skip whitespace          */

    /* ------------( arithmetic operators )------------------------- */

"+"                     return PLUS;            /* arith. addition          */
"-"                     return MINUS;           /* arith. subtraction       */
"*"                     return MULT;            /* arith. multiplication    */
"/"                     return DIV;             /* arith. division          */
"**"                    return EXPO;            /* arith. exponentation     */
"%"                     return MODULO;          /* arith. modulo            */

    /* --------------------( logical operators )---------------------*/
"&&"|"."{a}{n}{d}"."    return ANDIF;           /* logical and              */
"||"|"."{o}{r}"."       return ORIF;            /* logical or               */
"!"|"."{n}{o}{t}"."     return NOT;             /* logical not              */
{true}|"."{true}"."         return TRUECONST;       /* logical true             */


    /* --------------------( bitwise operators )---------------------*/
"&"                     return AND;             /* bitwise and              */
"|"                     return OR;              /* bitwise or               */
"^"                     return XOR;             /* exclusive or             */
"<<"                    return LSHIFT;          /* bitshift left            */
">>"                    return RSHIFT;          /* bitshift right           */
"~"                     return BITCOMPL;        /* one's complement         */

    /* ------------------( relational operators )--------------------*/
"<"|"."{l}{t}"."        { DssParser::yylval.yRelOp = DssParser::LtOpT; return RELOP; }
"<="|"."{l}{e}"."       { DssParser::yylval.yRelOp = DssParser::LeOpT; return RELOP; }
">"|"."{g}{t}"."        { DssParser::yylval.yRelOp = DssParser::GtOpT; return RELOP; }
">="|"."{g}{e}"."       { DssParser::yylval.yRelOp = DssParser::GeOpT; return RELOP; }
"=="|"."{e}{q}"."       { DssParser::yylval.yRelOp = DssParser::EqOpT; return RELOP; }
"!="|"."{n}{e}"."       { DssParser::yylval.yRelOp = DssParser::NeOpT; return RELOP; }

    /* ------------------( arithmetic functions )---------------------*/
{fabs}/{wpar}           { DssParser::yylval.yFunc1 = fabs; return AFUNC1; }

{fsin}/{wpar}           { DssParser::yylval.yFunc1 = sin; return AFUNC1; }
{fcos}/{wpar}           { DssParser::yylval.yFunc1 = cos; return AFUNC1; }
{ftan}/{wpar}           { DssParser::yylval.yFunc1 = tan; return AFUNC1; }
{asin}/{wpar}           { DssParser::yylval.yFunc1 = asin; return AFUNC1; }
{acos}/{wpar}           { DssParser::yylval.yFunc1 = acos; return AFUNC1; }
{atan2}/{wpar}          { DssParser::yylval.yFunc2 = atan2; return AFUNC2; }
{atan}/{wpar}           { DssParser::yylval.yFunc1 = atan; return AFUNC1; }

{sinh}/{wpar}           { DssParser::yylval.yFunc1 = sinh; return AFUNC1; }
{cosh}/{wpar}           { DssParser::yylval.yFunc1 = cosh; return AFUNC1; }
{tanh}/{wpar}           { DssParser::yylval.yFunc1 = tanh; return AFUNC1; }

{fexp}/{wpar}           { DssParser::yylval.yFunc1 = exp; return AFUNC1; }
{flog}/{wpar}           { DssParser::yylval.yFunc1 = log; return AFUNC1; }
{flog10}/{wpar}         { DssParser::yylval.yFunc1 = log10; return AFUNC1; }
{fsqrt}/{wpar}          { DssParser::yylval.yFunc1 = sqrt; return AFUNC1; }

{modf}/{wpar}           { DssParser::yylval.yFunc1 = DssMisc::modf;
                          return AFUNC1; }
{fint}/{wpar}           { DssParser::yylval.yFunc1 = DssMisc::intf;
                          return AFUNC1; }
{pow}/{wpar}            { DssParser::yylval.yFunc2 = pow; return AFUNC2; }
{ceil}/{wpar}           { DssParser::yylval.yFunc1 = ceil; return AFUNC1; }
{floor}/{wpar}          { DssParser::yylval.yFunc1 = floor; return AFUNC1; }
{fmod}/{wpar}           { DssParser::yylval.yFunc2 = fmod; return AFUNC2; }

    /* -------------------( region shape functions )----------------------*/
{point}/{wpar}          { return RPOINT; }
{line}/{wpar}           { return RLINE; }
{circle}/{wpar}         { return RCIRCLE; }
{ring}/{wpar}           { return RRING; }
{ellipse}/{wpar}        { return RELLIPSE; }
{elliptr}/{wpar}        { return RELLIPSERING; }
{box}/{wpar}            { return RBOX; }
{rect}/{wpar}           { return RRECTANGLE; }
{polygon}/{wpar}        { return RPOLYGON; }
{sector}/{wpar}         { return RSECTOR; }
{diamond}/{wpar}        { return RDIAMOND; }

    /* ---------------------(block filters )------------------------------*/
{gti}/"("               { BEGIN EXPECTBLOCKNAME; return BLOCKFILTER1; }
{mask}/"("              { BEGIN EXPECTBLOCKNAME; return BLOCKFILTER2; }
{region}/"("            { BEGIN EXPECTBLOCKNAME; return BLOCKFILTER2; }
<EXPECTBLOCKNAME>{blockname} { BEGIN INITIAL;
                               DssParser::yylval.yBlockSpec = new std::string(yytext);
                               return BLOCK; }

    /* -----------------------( numbers )--------------------------------*/
{binnum}                { DssParser::yylval.yNumber = DBL(
                                strtoul(yytext+1, // skip the 'b'
                                (char **)NULL, 2)); return NUMBER; }

{octnum}                { DssParser::yylval.yNumber = DBL(
                                strtoul(yytext+1, // skip the preceding '0'
                                (char **)NULL, 8)); return NUMBER; }

{decnum}                { DssParser::yylval.yNumber = DBL(
                                strtoul(yytext, (char **)NULL, 10));
                          return NUMBER; }

{hexnum}                { DssParser::yylval.yNumber = DBL(
                                strtoul(yytext+2, // skip the preceing '0x'
                                (char **)NULL, 16)); return NUMBER; }

{realnum}               { DssParser::yylval.yNumber = atof(yytext);
                          return NUMBER; }

"#"{p}{i}               { DssParser::yylval.yNumber = M_PI; return NUMBER; }
"#"{e}                  { DssParser::yylval.yNumber = M_E;  return NUMBER; }
"#"{d}{e}{g}            { DssParser::yylval.yNumber = DEG_TO_RAD;
                          return NUMBER; }
"#"{r}{a}{d}            { DssParser::yylval.yNumber = RAD_TO_DEG;
                          return NUMBER; }

{alpha}{alnum}*         { DssParser::yylval.yAxisName = new std::string(yytext);
                          return AXISNAME; }

.                       { return yytext[0]; }

%%

/*
 *  line I/O for the scanner
 */
static char *buf, *bp;

namespace DssParser {

extern const Table * yyFilteredTable;

const Table * yyFilteredTable;

void lexinit(const std::string &expression, const Table * filteredTable)
{
    YY_FLUSH_BUFFER;
    yyFilteredTable  = filteredTable;
    buf = bp = const_cast<char *>(expression.c_str());
#ifdef  FLEX_SCANNER
    yy_set_bol(1)
    BEGIN(INITIAL);
#endif
}

#ifdef  FLEX_SCANNER

static int lexiread (char *line, int max)
{   int n;

    for (n=0; *bp && n<max; ++n)
        *line++ = *bp++;

    return n;
}

//
//  declaration of dum1/dum2 is workaround to avoid compiler warning
//  about yy_flex_realloc/yyunput being defined but not used
//  should be replaced by compiler pragma to turn that message of
//
void *(*dum1)(void *, unsigned) = yy_flex_realloc;
void (*dum2)(int, char *) = yyunput;

#else

static int input (void)
{
    return * bp ? * bp ++ : * bp;
}

static int unput (int ch)
{
    return ch ? (*--bp = ch) : (*bp = ch);
}

static int output (int ch)
{
    assert(0);
}

#endif  /* FLEX_SCANNER */

void yyerror(char *s)
{
    using namespace SelectError;

    selerrstr<<buf<<endl;
    for (char *cp = buf; cp < bp; ++cp)
        selerrstr<<(*cp == '\t' ? '\t' : ' ');
    selerrstr<<"^ "<<s<<endl;
}

}
