//
//  Definition of the DataSubspace
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/


#include <list>
#include <Dal.h>
#include <Region.h>

#ifndef DSS_H
#define DSS_H

template<class T> class num_limits {public: static T max;};

class DScomp;
class DSfilter;
class DataSubspace;
template<class T> class Range;


enum RangeBound {Open=1, Closed, Unbounded};


template<class T> class DSchild{
public:
  DSchild( const T * parent ) : _parent( parent ) {}
  T * parent() { return const_cast<T *>(_parent); }
  const T * parent() const { return _parent; }
  void parent( const T * parent ) { _parent = parent;}
  bool isOrphan() const { return _parent == 0; }

private :
  const T * _parent;
};


class DSS {
public:
  typedef std::vector<std::string> DSfilterComp;
  static const DSfilterComp noComps;

  typedef std::list<std::string> DSaxes;
  static const DSaxes noAxes;

  //  The available filter types.
  enum FilterType {RangeT = 1, RegionT, BitMaskT, RangeBlockT,
                   GtiT, MaskT, RegionBlockT};
};


class DataSubspace : public DSS {
public:
  DataSubspace();

  // Explicit copy and initialization
  DataSubspace(const DataSubspace &dss);
  DataSubspace &operator=(const DataSubspace &dss);

  //  This constructor also ingests any DataSubspace info already
  //  in the Block.
  DataSubspace(Block *inBlock);

  //  A DS is made up of components.  Each component specifies a
  //  volume in data space.  It can also be thought of as a group
  //  of ANDed filters.
  DScomp *addComponent();

  // Add a copy of a component
  DScomp *addComponent(const DScomp &dscomp);

  void deleteComponent(unsigned compNumber);

  DScomp *component(unsigned compNumber) const;

  //  Number of components
  unsigned components() const;

  //  Return a string containing a valid selectlib expression
  //  for this data subspace, projected onto the input axes.
  std::string selectExpression(const DSaxes &axes=noAxes) const;

  //  Calculate an N-dimensional volume for the listed axes of the
  //  data subspace.
  double volume(const DSaxes &axes);

  // Return a list of ranges applicable to a given axis.  If the
  // the axis is not of type Range, then an empty list is returned.
  std::list<Range<double> > ranges(const std::string &axis) const;

  // What is the extent of a given axis, ie what are the min and
  // max values for this axis.  Return as a range.
  Range<double> extent(const std::string &axis);
  Range<double> realextent(const std::string &axis);

  // Does the subspace contain a filter 'like' the argument, ie it has the same
  // name, type, units and components
  bool hasFilter(const DSfilter &filter) const;

  // Does the subspace contain a filter with the given name and type
  bool hasFilter(const std::string &filterName,
		 const FilterType &filterType) const;

  // Is the filter global, ie does it appear in all components?
  bool isGlobal(const DSfilter &filter) const;

  // Where possible, combine together any GeoRegion filters
  void compressRegions();

  // Used to tell the outside world that there was a problem at some point
  void setError(bool error=true) {_error=error;}
  bool error() const{return _error;}

  // Negate a subspace, .ie take the logical not of the subspace.
  void operator!();

  ~DataSubspace();

private:
  // Vector holding the DSS components
  std::vector<DScomp*> _compvec;
  bool _error;
};


// operators to read and write DSS info from a block
Block *operator>>(Block *fromBlock, DataSubspace &inDss);
Block *operator<<(Block *toBlock, DataSubspace &outDss);

// Perform union of two DSS's.
DataSubspace &operator*=(DataSubspace &dss1, const DataSubspace &dss2);
DataSubspace operator*(const DataSubspace &dss1, const DataSubspace &dss2);

// Delete the DSS specification from a block, including referenced extensions.
void deleteDSS(Block *block);

// Clean the DSS specification.  Each component of the DSS is projected
// onto the table columns.  If the resulting selection does not accept any
// rows, then the component is deleted from the DSS.
void cleanDSS(const Table *tab, DataSubspace &dss);


class DScomp : public DSchild<DataSubspace>, public DSS {
  //  A DataSubspace component is just a list of filters which are
  //  implicitly ANDed together.  For most filter types, one can
  //  initialize them either 'by hand' or with a column, in which
  //  case it inherits all the quantities of the column.
public:
  DScomp(DataSubspace *parent);

  // Explicit copy and initialization
  DScomp(const DScomp &dscomp);
  DScomp &operator=(const DScomp &dscomp);

  // Add a copy of a filter
  DSfilter *addFilter(DSfilter &filter);

  //  Add a range filter
  DSfilter *addRangeFilter(const std::string &axisName,
		      const std::list<Range<long> > ranges,
		      const std::string &units = "");

  DSfilter *addRangeFilter(const std::string &axisName,
		      const std::list<Range<double> > ranges,
		      const std::string &units = "");

  DSfilter *addRangeFilter(const Column *inheritedColumn,
		      const std::list<Range<long> > ranges);

  DSfilter *addRangeFilter(const Column *inheritedColumn,
		      const std::list<Range<double> > ranges);

  // Add a bit mask filter.
  DSfilter *addBitFilter(const std::string &axisName,
			 std::string bitMaskString);

  DSfilter *addBitFilter(const Column *inheritedColumn,
			 std::string bitMaskString);

  DSfilter *addBitFilter(const std::string &axisName,
			 vector<uint32_t> onBitMasks,
			 vector<uint32_t> offBitMasks);

  DSfilter *addBitFilter(const Column *inheritedColumn,
			 vector<uint32_t> onBitMasks,
			 vector<uint32_t> offBitMasks);

  //  Add a block filter (either image or table).  The
  //  Filter Component is required for 2D filters, but not all
  //  table filters are necessarily 2D.
  DSfilter *addBlockFilter(const std::string &axisName,
			   const std::string &blockSpec,
			   const std::string &units = "");

  DSfilter *addBlockFilter(const Column *inheritedColumn,
			   const std::string &blockSpec);

  DSfilter *addBlockFilter(const std::string &axisName,
			   const std::string &blockSpec,
			   DSfilterComp &filtercomp);

  //  Add a coordinate region filter.  THis is based on a convention
  //  for the name from [1], section 2.2.  For example, if name='POS',
  //  then there should be two filter components, called 'X' and 'Y'
  //  with WCS information.  This convention can also be used above
  //  with block filters.
  DSfilter *addRegionFilter(const std::string &axisName,
			    GeoRegion region);

  DSfilter *addRegionFilter(const std::string &axisName,
			    GeoRegion region,
			    DSfilterComp &filtercomp);

  void deleteFilter(unsigned filterNumber);

  DSfilter *filter(unsigned filterNumber) const;

  DSfilter *filter(const std::string &filterName,
		   const FilterType &filterType) const;

  // Does the component have a filter with the given name and type
  bool hasFilter(const std::string &filterName,
		 const FilterType &filterType) const;

  // Return a filter that is 'like' the argument, ie it has the same
  // name, type, units and components
  DSfilter *filter(const DSfilter &filter) const;

  // Does the component contain a filter 'like' the one specified
  bool hasFilter(const DSfilter &filter) const;

  //  Number of filters in this component.
  unsigned filters() const;

  //  Return a string containing a valid selectlib expression
  //  for this component, projected onto the input axes.
  std::string selectExpression(const DSaxes &axes=noAxes) const;

  //  Calculate an N-dimensional volume for the listed axes of the
  //  data component.
  double volume(const DSaxes &axes);

  // Return a list of ranges applicable to a given axis.  If the
  // the axis is not of type Range, then an empty list is returned.
  std::list<Range<double> > ranges(const std::string &axis) const;

  // What is the extent of a given axis, ie what are the min and
  // max values for this axis.  Return as a range.
  Range<double> extent(const std::string &axis);
  Range<double> realextent(const std::string &axis);

  // Perform union of two subspace components.
  DScomp &operator*=(const DScomp &dss);

  // Where possible, combine together any GeoRegion filters
  void compressRegions();

  // Does the component contain a null filter
  bool isNull() const {return _null;}

  ~DScomp();
  /* AI gcc4 porting
   * extra qualification ‘DScomp::’ on member ‘_nameCheck’ => remove qualification
   * ‘void DScomp::_nameCheck(DSfilter&)’ is private => made it public...
   */
  // Resolves name conflicts between filters
  void _nameCheck(DSfilter &filter);

 private:
  // Vector holding the DSS filters
  std::vector<DSfilter *> _filtvec;


  // Fix block specification to conform to Dal standard
  std::string _fixBlockSpec(const std::string &blockSpec);

  // Does the component contain a filter which selects nothing
  bool _null;

};


// Perform union of two subspace components.
DScomp operator*(const DScomp &dscomp1, const DScomp &dscomp2);


class DSfilter : public DSchild<DScomp>, public DSS {
  //  Base class for the various kinds of filters
 public:
  // Root for DSS comment string
  static const std::string comment;

  // Generic constructors
  DSfilter(DScomp *parent,
	   const FilterType ftype, const std::string &name,
	   const std::string &units = "",
	   const DSfilterComp &filtercomp = noComps);

  DSfilter(DScomp *parent,
	   const FilterType ftype,
	   const Column *inheritedColumn);

  FilterType type() const;
  std::string name() const;
  std::string units() const;
  DSfilterComp comps() const;

  //  Return a string containing a valid selectlib expression
  //  for this filter.
  virtual std::string selectExpression() = 0;

  //  Calculate an N-dimensional volume for the component(s) of this
  //  filter.
  virtual double volume() = 0;

  // Write out the filter to a Block, using the component and
  // filter number specified.
  virtual void write(unsigned compNumber, unsigned filterNumber,
		     Block *toBlock, bool dummy=false) = 0;

  // Are two filters just copies of each other (or not).
  virtual bool operator==(DSfilter &) const = 0;

  bool operator!=(DSfilter & oFilter) const{ return !(*this==oFilter); }

  // Are two filter 'like' each other, ie. same name, type, units and components
  bool like(const DSfilter &) const;

  // What is the extent of the filter axis, ie what are the min and
  // max values for this axis.  Return as a range.
  virtual Range<double> extent(const std::string &axis);
  virtual Range<double> realextent(const std::string &axis);

  // Negate a filter, .ie take the logical not of the filter.
  virtual void operator!();

  //  Methods for accessing data in the various types of filters.

  virtual std::list<Range<double> > ranges() const;

  virtual vector<uint32_t> onBitMasks() const;
  virtual vector<uint32_t> offBitMasks() const;

  virtual GeoRegion region() const;

  virtual std::string blockSpec() const;

  virtual ~DSfilter() {}

 protected:
  FilterType _ftype;
  std::string _name;
  std::string _units;
  DSfilterComp _comps;
  // The identifier is used to distinguish between two filters within a
  // component, and to link filters between components.
  unsigned _identifier;

  void _getRefBlockName(unsigned compNumber, unsigned filterNumber,
			Block *toBlock, std::string blockSpec, std::string &ref) const;

  GeoRegion _negateRegion(GeoRegion &region);

  friend void DScomp::_nameCheck(DSfilter &filter);
};


template<class T> class DSrangeFilter : public DSfilter {
public:
  //  Initialize a filter with a range specification
  DSrangeFilter(DScomp *parent,
		const std::string &name,
		const std::list<Range<T> > range,
		const std::string &units);

  DSrangeFilter(DScomp *parent,
		const Column *inheritedColumn,
		const std::list<Range<T> >);

  // extent (min and max range) of the filter axis
  Range<double> extent(const std::string &axis);
  Range<double> realextent(const std::string &axis);

  //  Access to the range information
  std::list<Range<double> > ranges() const;

  // Return a string containing a valid selectlib expression
  //  for this filter.
  std::string selectExpression();

  //  Calculate an N-dimensional volume for the component(s) of this
  //  filter.
  virtual double volume() = 0;

  // Write out the filter to a Block, using the component and
  // filter number specified.
  void write(unsigned compNumber, unsigned filterNumber,
	     Block *toBlock, bool dummy=false);

  // Or two filters just copies of each other (or not).
  bool operator==(DSfilter &) const;

  // Negate a filter, .ie take the logical not of the filter.
  void operator!();

protected:
  std::list<Range<T> > _ranges;
};


template<class T>
class DSintRange : public DSrangeFilter<T> {
 public:
  //  Initialize a filter with a range specification
  DSintRange(DScomp *parent,
	     const std::string &name,
             const std::list<Range<T> > range,
	     const std::string &units);

  DSintRange(DScomp *parent,
	     const Column *inheritedColumn,
	     const std::list<Range<T> > range);

  //  Calculate an N-dimensional volume for the component(s) of this
  //  filter.
  double volume();
};


template<class T>
class DSrealRange : public DSrangeFilter<T> {
 public:
  //  Initialize a filter with a range specification
  DSrealRange(DScomp *parent,
	      const std::string &name,
	      const std::list<Range<T> > range,
	      const std::string &units);

  DSrealRange(DScomp *parent,
	      const Column *inheritedColumn,
	      const std::list<Range<T> >);

  //  Calculate an N-dimensional volume for the component(s) of this
  //  filter.
  double volume();
};


class DSbitFilter : public DSfilter {
public:
  //  Initialize the filter with a bit mask.
  DSbitFilter(DScomp *parent,
	      const std::string &name,
	      std::string bitString);

  DSbitFilter(DScomp *parent,
	      const Column *inheritedColumn,
	      std::string bitString);

  DSbitFilter(DScomp *parent,
	      const std::string &name,
	      vector<uint32_t> onBitMasks,
	      vector<uint32_t> offBitMasks);

  DSbitFilter(DScomp *parent,
	      const Column *inheritedColumn,
	      vector<uint32_t> onBitMasks,
	      vector<uint32_t> offBitMasks);

  //  Return the value of the bitMask
  vector<uint32_t> onBitMasks() const;
  vector<uint32_t> offBitMasks() const;

  // Return a string containing a valid selectlib expression
  //  for this filter.
  std::string selectExpression();

  //  Calculate an N-dimensional volume for the component(s) of this
  //  filter.
  double volume();

  // Write out the filter to a Block, using the component and
  // filter number specified.
  void write(unsigned compNumber, unsigned filterNumber,
	     Block *toBlock, bool dummy=false);

  // Or two filters just copies of each other (or not).
  bool operator==(DSfilter &) const;

  // Negate a filter, .ie take the logical not of the filter.
  void operator!();

private:
  void _parseBitString(std::string bitString);
  vector <uint32_t> _onBits, _offBits;
};


class DSblockFilter : public DSfilter {
public:
  // explicit copy
  DSblockFilter(DSblockFilter &dsblock);

  //  Initialize the filter with a block pointer
  DSblockFilter(DScomp *parent,
		const FilterType ftype,
		const std::string &name,
		const std::string &blockSpec,
		const std::string &units = "");

  DSblockFilter(DScomp *parent,
		const FilterType ftype,
		const Column *inheritedColumn,
		const std::string &blockSpec);

  DSblockFilter(DScomp *parent,
		const FilterType ftype,
		const std::string &name,
		const std::string &blockSpec,
		DSfilterComp &filtercomp);

  //  Return a string containing the block specification
  std::string blockSpec() const;

  //  Return a string containing a valid selectlib expression
  //  for this filter.
  virtual std::string selectExpression()=0;

  //  Calculate an N-dimensional volume for the component(s) of this
  //  filter.
  virtual double volume() = 0;

  // Write out the filter to a Block, using the component and
  // filter number specified.
  virtual void write(unsigned compNumber, unsigned filterNumber,
		     Block *toBlock, bool dummy=false);

  // Are two filters just copies of each other (or not).
  bool operator==(DSfilter &) const;

  ~DSblockFilter();

protected:
  void _openBlock();
  void _setBlockSpec(const std::string &blockSpec);
  Block * _block;
  std::string _parentName;
  std::string _blockName;
};



class DSrangeBlock : public DSblockFilter {
 public:
  // explicit copy
  DSrangeBlock(DSrangeBlock &dsblock);

  DSrangeBlock(DScomp *parent,
	       const std::string &name,
	       const std::string &blockSpec,
	       const std::string &units = "");

  DSrangeBlock(DScomp *parent,
	       const Column *inheritedColumn,
	       const std::string &blockSpec);

  //  Return a string containing a valid selectlib expression
  //  for this filter.
  std::string selectExpression();

  // extent (min and max range) of the filter axis
  Range<double> extent(const std::string &axis);

  //  Calculate an N-dimensional volume for the component(s) of this
  //  filter.
  double volume();
};



class DSgtiBlock : public DSblockFilter {
 public:
  // explicit copy
  DSgtiBlock(DSgtiBlock &dsblock);

  DSgtiBlock(DScomp *parent,
	     const std::string &name,
	     const std::string &blockSpec,
	     const std::string &units = "");

  DSgtiBlock(DScomp *parent,
	     const Column *inheritedColumn,
	     const std::string &blockSpec);

  // extent (min and max range) of the filter axis
  Range<double> extent(const std::string &axis);

  //  Return a string containing a valid selectlib expression
  //  for this filter.
  std::string selectExpression();

  // Write out the filter to a Block, using the component and
  // filter number specified.
  void write(unsigned compNumber, unsigned filterNumber,
	     Block *toBlock, bool dummy=false);

  // Negate a filter, .ie take the logical not of the filter.
  void operator!();

  //  Calculate an N-dimensional volume for the component(s) of this
  //  filter.
  double volume();
};



class DSregionBlock : public DSblockFilter {
 public:
  // explicit copy
  DSregionBlock(DSregionBlock &dsblock);

  DSregionBlock(DScomp *parent,
		const std::string &name,
		const std::string &blockSpec,
		DSfilterComp &filtercomp);

  DSregionBlock(DScomp *parent,
		const std::string &name,
		const std::string &blockSpec);

  // extent (min and max range) of the filter axis
  Range<double> extent(const std::string &axis);
  Range<double> realextent(const std::string &axis);

  //  Return a string containing a valid selectlib expression
  //  for this filter.
  std::string selectExpression();

  //  Calculate an N-dimensional volume for the component(s) of this
  //  filter.
  double volume();

  //  Return a corresponding region object
  GeoRegion region() const;

  // Negate a filter, .ie take the logical not of the filter.
  void operator!();
};



class DSmaskBlock : public DSblockFilter {
public:
  // explicit copy
  DSmaskBlock(DSmaskBlock &dsblock);

  DSmaskBlock(DScomp *parent,
	      const std::string &name,
	      const std::string &blockSpec,
	      DSfilterComp &filtercomp);

  DSmaskBlock(DScomp *parent,
	      const std::string &name,
	      const std::string &blockSpec);

  //  Return a string containing a valid selectlib expression
  //  for this filter.
  std::string selectExpression();

  //  Calculate an N-dimensional volume for the component(s) of this
  //  filter.
  double volume();
};


class DSregionFilter : public DSfilter {
public:
  //  Initialize the filter with a Region object
  DSregionFilter(DScomp *parent,
		 const std::string &name,
		 GeoRegion region,
		 DSfilterComp &filtercomp);

  DSregionFilter(DScomp *parent,
		 const std::string &name,
		 GeoRegion region);

  //  Return a reference to the region object
  GeoRegion region() const;

  // extent (min and max range) of the filter axis
  Range<double> extent(const std::string &axis);
  Range<double> realextent(const std::string &axis);

  //  Return a string containing a valid selectlib expression
  //  for this filter.
  std::string selectExpression();

  //  Calculate an N-dimensional volume for the component(s) of this
  //  filter.
  double volume();

  // Write out the filter to a Block, using the component and
  // filter number specified.
  void write(unsigned compNumber, unsigned filterNumber,
	     Block *toBlock, bool dummy=false);

  // Negate a filter, .ie take the logical not of the filter.
  void operator!();

  // Are two filters just copies of each other (or not).
  bool operator==(DSfilter &) const;

private:
  GeoRegion _region;
};


template<class T> class Range{
public:
  //  A range is a type that has an upper and lower limit,
  //  both of which are designated as either open or closed, or undefined.

  Range(bool nullify=false);

  template<class U> Range(const Range<U> &);

  Range(T lower, T upper,
	RangeBound lowerB=Closed, RangeBound upperB=Closed);

  template<class U> Range(U lower, U upper,
			  RangeBound lowerB=Closed, RangeBound upperB=Closed);

  void values(T &lower, T &upper,
	      RangeBound &lowerB, RangeBound &upperB) const;

  bool contains(T value) const;

  bool isNull() {return _null;}

  template<class U> bool operator<(const Range<U>&) const;
  template<class U> bool operator==(const Range<U>&) const;
  template<class U> bool operator!=(const Range<U>&) const;

private:
  T _lower, _upper;
  RangeBound _lowerB, _upperB;
  bool _null;
};

// Utility function to parse a string containing an ASC range specification

template<class T>
std::list<Range<T> > parseRanges(const std::string &rangeSpec);

template<class T>
Range<T> maxRange(const std::list<Range<T> > &ranges);

template<class T>
std::list<Range<T> > operator&(std::list<Range<T> > rl1, std::list<Range<T> > rl2);

// Utility functions for GeoRegion's

void geoExtent(const GeoRegion &region, int &Xmin, int &Xmax, int &Ymin, int &Ymax);
void realgeoExtent(const GeoRegion &region, double &Xmin, double &Xmax, double &Ymin, double &Ymax);
unsigned long geoArea(const GeoRegion &region);

#endif
